@@ -1,4 +1,4 @@
-from .structure_learning import hc, pc_stable, ges
+from .structure_learning import hc, pc_stable, ges, sed
 from .evaluation import f1, shd
 from .accessory import miss_mechanism, add_missing, dag2cpdag, from_bnlearn, to_bnlearn, pdag2dag, edges, parents, \
     neighbours, g_test, pearson_test
