@@ -9,7 +9,7 @@ import numpy as np
 from graphviz import Digraph
 from numba import njit
 from scipy.stats.distributions import chi2
-from .score import compute_weights
+from scipy import stats
 
 
 # random.seed(941214)
@@ -122,8 +122,9 @@ def add_missing(data, cause_dict, m_min=0, m_max=0.5):
                         states.remove(state_single)
                         if data[cause].value_counts()[selected_state].sum() > m * data.shape[0] or len(states) == 1:
                             m_h = np.random.uniform(m, m * data.shape[0] / data[cause].isin(selected_state).sum())
-                            m_l = max((data.shape[0] * m - data[cause].isin(selected_state).sum() * m_h) / data[cause].isin(
-                                states).sum(), 0)
+                            m_l = max(
+                                (data.shape[0] * m - data[cause].isin(selected_state).sum() * m_h) / data[cause].isin(
+                                    states).sum(), 0)
                             break
                     data_missing[var][
                         (np.random.uniform(size=len(data)) < m_h) & (data[cause].isin(selected_state))] = None
@@ -139,6 +140,58 @@ def add_missing(data, cause_dict, m_min=0, m_max=0.5):
     return data_missing
 
 
+@njit(fastmath=True)
+def density_counter(data, arities):
+    strides = np.empty(data.shape[1], dtype=np.uint32)
+    idx = data.shape[1] - 1
+    stride = 1
+    while idx > -1:
+        strides[idx] = stride
+        stride *= arities[idx]
+        idx -= 1
+    CT = np.zeros(stride)
+    for rowidx in range(data.shape[0]):
+        idx = 0
+        for i in range(data.shape[1]):
+            idx += data[rowidx, i] * strides[i]
+        CT[idx] += 1
+
+    CT /= np.sum(CT)
+    return CT
+
+
+# compute weights of data with missing values
+def compute_weights(data, varnames, W_ids, cause_list):
+    if 'int' in data.dtype.name:
+        data_delete = data[(data[:, W_ids].min(axis=1) >= 0) & (data[:, W_ids].max(axis=1) != np.iinfo(data.dtype).max)]
+    else:
+        data_delete = data[~np.isnan(data[:, W_ids]).any(axis=1)]
+    weights = np.ones(len(data_delete))
+    for ri in [varnames[i] for i in W_ids]:
+        cause = cause_list[ri]
+        if len(cause) > 0:
+            cause_id = [i for i in range(len(varnames)) if varnames[i] in cause]
+            data_beta = data[:, cause_id]
+            if 'int' in data.dtype.name:
+                numerator = data_beta[
+                    (data_beta.min(axis=1) >= 0) & (data_beta.max(axis=1) != np.iinfo(data.dtype).max)]
+                denominator = data_beta[
+                    (data_beta.min(axis=1) >= 0) & (data_beta.max(axis=1) != np.iinfo(data.dtype).max) & (
+                            data[:, varnames.index(ri)] >= 0)]
+                arities = np.amax(numerator, axis=0) + 1
+                f_w = density_counter(numerator, arities).reshape(arities)
+                f_wr = density_counter(denominator, arities).reshape(arities)
+                weights *= np.array([f_w[tuple(data_delete[i, cause_id])] for i in range(len(data_delete))]) / np.array(
+                    [f_wr[tuple(data_delete[i, cause_id])] for i in range(len(data_delete))])
+            else:
+                numerator = data_beta[~np.isnan(data_beta).any(axis=1)].T
+                denominator = data_beta[(~np.isnan(data_beta).any(axis=1)) & (~np.isnan(data[:, varnames.index(ri)]))].T
+                f_w = stats.gaussian_kde(numerator)
+                f_wr = stats.gaussian_kde(denominator)
+                weights *= f_w(data_delete[:, cause_id].T) / f_wr(data_delete[:, cause_id].T)
+    return weights * len(weights) / weights.sum()
+
+
 def pairwise(data, varnames, vars, cause_list, cache_data, cache_weight, method):
     if method == 'pw':
         W = tuple(sorted([v for v in vars if v in cause_list]))
@@ -226,7 +279,7 @@ def find_causes(data, test_function='default', alpha=0.01):
                 for cond in combinations(cond_set, l):
                     data_delete = data[(data[[can] + list(cond)] > -1).all(1)].to_numpy() if factor else data.dropna(
                         subset=[can] + list(cond))
-                    if data_delete.shape[0] > 0:
+                    if data_delete.shape[0] > 0 and data_delete.iloc[:, -1].unique().shape[0] > 1:
                         cols = np.asarray(
                             [data_delete.shape[1] - 1] + [varnames.index(can)] + [varnames.index(x) for x in cond])
                         p_value = globals()[test_function](data_delete, cols)
@@ -487,9 +540,12 @@ def g_counter(data, arities, cols, weights=None):
 def pearson_test(data, cols):
     data = pandas.DataFrame(data)
     if len(cols) == 2:
-        return pg.partial_corr(data, cols[0], cols[1])['p-val'][0]
+        return pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]])['p-val'][0]
     elif len(cols) > 2:
-        return pg.partial_corr(data, cols[0], cols[1], cols[2:])['p-val'][0]
+        return \
+            pg.partial_corr(data, data.columns[cols[0]], data.columns[cols[1]], data.columns[cols[2:]].to_list())[
+                'p-val'][
+                0]
     else:
         raise Exception('Length of input cols is less than 2')
 
@@ -641,7 +697,7 @@ def dag2cpdag(dag):
 
 
 # orient v-structrues
-def orient_v_structure(data, varnames, pc, sepset, ci_test, threshold=0.05):
+def orient_v_structure(data, varnames, pc, sepset, ci_test):
     dag = {}
     v_candidate = {}
     for var in pc:
@@ -757,3 +813,83 @@ def edges(graph):
             if len(edge[(edge['from'] == nei) & (edge['to'] == var)]) == 0:
                 edge.loc[len(edge)] = [var, nei, 'undirected']
     return edge
+
+
+def pre_process(data):
+    '''
+    preprocess the original input data to remove unused states and single state columns
+    :param data: input data in pandas dataframe format
+    :return: prunned data and list of single value columns
+    '''
+    # remove unused states
+    if all(data[var].dtype.name == 'category' for var in data):
+        for var in list(data):
+            data[var] = data[var].cat.remove_unused_categories()
+        # data = pandas.DataFrame(data.to_numpy(), columns=list(data.columns), dtype='category')
+    # remove single value columns
+    dc = data.nunique() == 1
+    dc = dc[dc].index.values
+    data = data.drop(dc, axis=1)
+    return data, dc
+
+
+def post_process(graph, dc):
+    '''
+    postprocess the learned graph to add those single value variables
+    :param graph: learned graph
+    :param dc: list of single value columns
+    :return: expanded graph
+    '''
+    for node in dc:
+        graph[node] = {'par': [], 'nei': []}
+    return graph
+
+
+def check_cycle(vi, vj, dag):
+    # whether adding or orientating edge vi->vj would cause cycle. In other words, this function check whether there is a direct path from vj to vi except the possible edge vi<-vj
+    underchecked = [x for x in dag[vi]['par'] if x != vj]
+    checked = []
+    cyc_flag = False
+    while underchecked:
+        if cyc_flag:
+            break
+        underchecked_copy = list(underchecked)
+        for vk in underchecked_copy:
+            if dag[vk]['par']:
+                if vj in dag[vk]['par']:
+                    cyc_flag = True
+                    break
+                else:
+                    for key in dag[vk]['par']:
+                        if key not in checked + underchecked:
+                            underchecked.append(key)
+            underchecked.remove(vk)
+            checked.append(vk)
+    return cyc_flag
+
+
+def drop_edge(dag, edge):
+    dag_dropped = deepcopy(dag)
+    if edge[0] in dag_dropped[edge[1]]['par']:
+        dag_dropped[edge[1]]['par'].remove(edge[0])
+    elif edge[1] in dag_dropped[edge[0]]['par']:
+        dag_dropped[edge[0]]['par'].remove(edge[1])
+    else:
+        dag_dropped[edge[0]]['nei'].remove(edge[1])
+        dag_dropped[edge[1]]['nei'].remove(edge[0])
+    return dag_dropped
+
+
+# get markov blanket of node
+def markov_blanket(dag, var):
+    mb = deepcopy(dag[var]['par'])
+    if var == 'VENTALV':
+        a = 1
+    for v in dag:
+        if var in dag[v]['par']:
+            if v not in mb:
+                mb.append(v)
+            for par in dag[v]['par']:
+                if par != var and par not in mb:
+                    mb.append(par)
+    return mb
