@@ -957,8 +957,8 @@ def sed(learned_dag, data):
     '''
     if type(learned_dag) is str:
         learned_dag = from_bnlearn(learned_dag)
-    elif type(graph) is not dict:
-        raise Exception('The format of input graph is invalid.')
+    elif type(learned_dag) is not dict:
+        raise Exception('The format of input DAG is invalid.')
     # remove unused states
     data, _ = pre_process(data)
     varnames = data.columns.to_list()
@@ -1015,6 +1015,6 @@ def sed(learned_dag, data):
                 break
     dag_consistent = pdag2dag(corrected_cpdag)
     if dag_consistent is None:
-        return graph
+        return learned_dag
     else:
         return to_bnlearn(dag_consistent)
